#!/usr/bin/env bash

DOCOPT_PROGRAM_VERSION='1.5.2'

DOC="Naval Fate.

Usage:
  naval_fate.py ship new <name>...
  naval_fate.py ship <name> move <x> <y> [--speed=<kn>]
  naval_fate.py ship shoot <x> <y>
  naval_fate.py mine (set|remove) <x> <y> [--moored|--drifting]
  naval_fate.py -h | --help
  naval_fate.py --version

Options:
  -h --help     Show this screen.
  --version     Show version.
  --speed=<kn>  Speed in knots [default: 10].
  --moored      Moored (anchored) mine.
  --drifting    Drifting mine.
"
"DOCOPT PARAMS"
docopt "$@"

echo ${docopt_param_names[@]}
echo ${docopt_parsed_params[@]}
echo ${docopt_parsed_values[@]}
